/* heston.C: implementation for analytic solution of HESTON's Model
   for stochastic volatility
   Copyright (C) 2001 Gunter Winkler (guwi17@gmx.de)

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 
  author: Gunter Winkler (gw) 
          gunter.winkler@mathematik.tu-chemnitz.de
 
  paper:  'A closed-form solution for options with stochastic volatility'
          by Heston, L., 1995 Volatility/Risk Publications
 
  changes:
 	gw	2001-08-07	new
 	gw	2001-08-23	made public first time
 */

#include "heston.h"

#define Pi  3.14159265358979323846264338328
#define Pi2 1.57079632679489661923132169164

/* names: phi = 1: call, phi = -1: put */

/* some defines */
#define _2U(j)	((j==1)?(1):(-1))
#define _A  	(market.kappa * market.theta)
#define _B(j) 	((j==1)?(market.kappa+market.lambda-market.rho*market.xi):(market.kappa+market.lambda))

static heston_complex Fj( const int j,	/* j == 1 or j != 1 */
						  const DOUBLE x, 
						  const DOUBLE v,
						  const DOUBLE tau,
						  const DOUBLE phi,
						  const MARKETDATA market)
/*
	helper function f_j(x,v,tau;phi) from paper
*/
{
	heston_complex	C, D;	/* for C(tau,phi) and D(tau,phi) */
	heston_complex	g, d;	/* see paper */
	heston_complex  z1, z2; /* temporary */
	DOUBLE			rpx;	/* = rho * phi * xi */
	DOUBLE			xi2;	/* = xi^2 */

	DEBUG_HESTON(std::cout << "input: " << x << ", " << v << ", " << tau << ", " << phi << endl;)
		
	rpx = market.rho * market.xi * phi;
	xi2 = market.xi * market.xi;
	z1 = heston_complex( _B(j), -rpx );
	
	DEBUG_HESTON(std::cout << "rpx " << rpx << " xi^2 " << xi2 << " z1 " << z1 << endl;)

	DEBUG_HESTON(std::cout << "z1^2 " << (z1*z1) << " ... " << (xi2*(heston_complex(-phi*phi, _2U(j) * phi)));)
	d = z1*z1 - xi2 * heston_complex(-phi*phi, _2U(j) * phi);
	DEBUG_HESTON(std::cout << " d " << d;)
	d = sqrt( d );
	
	DEBUG_HESTON(std::cout << " d " << d;)
	
	z2 = (z1 + d);
	g = z2 / (z1 - d);
	
	DEBUG_HESTON(std::cout << " z2 " << z2;)
	DEBUG_HESTON(std::cout << " g " << g;)

	D = z2 / xi2 * ((DOUBLE)1.0-exp(d*tau)) / ((DOUBLE)1.0 - g*exp(d*tau));
	DEBUG_HESTON(std::cout << " D " << D;)
		
	C = heston_complex((DOUBLE)0.0,(market.rd-market.rf)*phi*tau)
			+ _A/xi2 * ( z2 * tau - (DOUBLE)2.0 * log(((DOUBLE)1.0-g*exp(d*tau))/((DOUBLE)1.0-g)) );
	DEBUG_HESTON(std::cout << " C " << C << endl;)
	
	return exp( C + D * v + heston_complex(0.0, phi*x) );
}


double	Pj ( const int j,
			 const DOUBLE x,
			 const DOUBLE v,
			 const DOUBLE tau,
			 const DOUBLE logstrike,
			 MARKETDATA market )
/*
	conditional probability P_j(x,v,tau;phi) from paper
*/
{
	const DOUBLE	phimin = 0.0;
	const DOUBLE	phimax = 100.0*Pi;
	const int		steps  = 100;
	const DOUBLE	stepsize = (phimax-phimin)/steps;
	
	const int		N = 3;
	
	const DOUBLE	x_i[N] = {	.112701665379258311482073460022,
								.5,
								.887298334620741688517926539978 };
	const DOUBLE	alpha_i[N] = {  .277777777777777777777777777778, 
									.444444444444444444444444444444, 
									.277777777777777777777777777778 };
		
	heston_complex	integrand;
	DOUBLE			phi, sum, zwsum;
	int 			i,k;
	
	sum = 0;
	for (i=steps-1;i>=0;--i)
	{
		for (k=N-1;k>=0;--k)
		{
			phi = (i + x_i[k]) * stepsize;
			integrand = (heston_complex) std::polar(1/phi,-phi*logstrike) 
					* (heston_complex(0,-1.0))
					* Fj(j,x,v,tau,phi,market);
			sum += alpha_i[k] * integrand.real();
			DEBUG_HESTON(std::cout << " phi " << phi << " int " << integrand << " sum " << sum << endl;)
		}
	}	
	sum *= stepsize;

	return 0.5 + sum / Pi;
}



double	Heston_Call( const DOUBLE spot,
					 const DOUBLE v,
					 const DOUBLE tau,
					 const DOUBLE strike,
					 MARKETDATA market )
/*
	formula for vanilla call (same as Black-Scholes)
*/
{
	DOUBLE	x, ls, p1, p2;
	
	x = log(spot);
	ls = log(strike);
	
	p1 = Pj(1, x, v, tau, ls, market);
	p2 = Pj(2, x, v, tau, ls, market);
	
	return spot * exp(-market.rf * tau) * p1 - strike * exp(-market.rd * tau) * p2;

}
